"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.AccessRightsService = void 0;
const tslib_1 = require("tslib");
/* eslint-disable @typescript-eslint/no-unused-vars */
const authorization_1 = require("@loopback/authorization");
const core_1 = require("@loopback/core");
const lodash_1 = require("lodash");
const application_error_1 = require("./../../helper/application-error");
const application_service_1 = require("./../application.service");
const object_node_service_1 = require("./../object-node.service");
const object_tree_service_1 = require("./../object-tree/object-tree.service");
const object_type_service_1 = require("./../object-type.service");
const access_rights_const_1 = require("./access-rights.const");
const access_rights_init_1 = require("./access-rights.init");
let AccessRightsService = class AccessRightsService {
    constructor(appCtx, objectTypeService, objectNodeService, objectTreeService) {
        this.appCtx = appCtx;
        this.objectTypeService = objectTypeService;
        this.objectNodeService = objectNodeService;
        this.objectTreeService = objectTreeService;
        this.accessRights = {};
        this.init = new access_rights_init_1.AccessRightsInit(appCtx, objectTypeService, objectNodeService);
    }
    get ready() {
        return this.init.ready;
    }
    async authorize(ctx, context, metadata) {
        try {
            await this.ready;
            ctx.accessRightsContexte.user.value =
                0 < context.principals.length
                    ? context.principals[0]
                    : null;
            if ((metadata === null || metadata === void 0 ? void 0 : metadata.resource) && (metadata === null || metadata === void 0 ? void 0 : metadata.resource) in this.accessRights) {
                const accessRights = this.accessRights[metadata.resource];
                if (metadata.scopes) {
                    if (context.scopes) {
                        context.scopes = context.scopes.concat(metadata.scopes);
                    }
                    else {
                        context.scopes = metadata.scopes;
                    }
                }
                return await accessRights.authorize(ctx, context);
            }
        }
        catch (error) {
            return authorization_1.AuthorizationDecision.DENY;
        }
        return authorization_1.AuthorizationDecision.ALLOW;
    }
    async cleanReturnedEntity(entityName, entity, ctx) {
        if (entityName in this.accessRights) {
            const accessRights = this.accessRights[entityName];
            await accessRights.cleanReturnedEntity(entity, ctx);
        }
        else {
            throw application_error_1.ApplicationError.forbiden();
        }
    }
    async cleanReturnedEntities(entityName, entities, ctx) {
        if (entityName in this.accessRights) {
            const accessRights = this.accessRights[entityName];
            for (let childIndex = entities.length - 1; childIndex >= 0; childIndex--) {
                // eslint-disable-next-line @typescript-eslint/no-explicit-any
                const entity = entities[childIndex];
                await this.cleanReturnedEntity(entityName, entity, ctx);
                if (entity.aclCtx && !entity.aclCtx.rights.read) {
                    entities.splice(childIndex, 1);
                }
            }
        }
    }
    registerAccessRightsService(resource, accessRights) {
        this.accessRights[resource] = accessRights;
    }
    async getNodeAccessRights(node, ctx) {
        // TODO : disable update for object created during boot ?
        try {
            const rights = await ctx.accessRightsContexte.rights.waitForValue;
            // if node is just the root of the acl tree
            if (node.id === ctx.accessRightsContexte.treeRootNodeId.value) {
                return rights.treeRootNode.toAccessRightCRUD();
            }
            // if owner => true
            if (rights.treeChildrenNodes.owner) {
                return new access_rights_const_1.AccessRightCRUD(true, true, true, true);
            }
            //Check if node is not in last def tree ACLDef
            //  => rights.treeChildrenNodes[scope]
            const aclTrees = await ctx.accessRightsContexte.aclTrees.waitForValue;
            if (!(node.parentACLId in aclTrees) ||
                !this.objectTreeService.isInTree(aclTrees[node.parentACLId], node)) {
                return rights.treeChildrenNodes.toAccessRightCRUD();
            }
            // if not acl => false (as the node is part of ACL)
            if (!rights.treeChildrenNodes.access) {
                return new access_rights_const_1.AccessRightCRUD(false, false, false, false);
            }
            // If Node = aclDef => read, write, create
            if (node.objectTypeId ===
                (await this.appCtx.accessRightsDefinitionType.waitForValue).id) {
                return new access_rights_const_1.AccessRightCRUD(true, true, false, true);
            }
            // If Node = Owners or children  => false
            const owners = this.objectTreeService.getChildOfType(aclTrees[node.parentACLId], (await this.appCtx.accessRightsOwnersType.waitForValue).id);
            if (owners && this.objectTreeService.isInTree(owners, node)) {
                return new access_rights_const_1.AccessRightCRUD(false, false, false, false);
            }
            // If Node = Managers or children => just read
            const managers = this.objectTreeService.getChildOfType(aclTrees[node.parentACLId], (await this.appCtx.accessRightsAccessManagersType.waitForValue)
                .id);
            if (managers && this.objectTreeService.isInTree(managers, node)) {
                return new access_rights_const_1.AccessRightCRUD(false, true, false, false);
            }
            // Else (acl and access right groups) => true
            return new access_rights_const_1.AccessRightCRUD(true, true, true, true);
        }
        catch (error) {
            return new access_rights_const_1.AccessRightCRUD(false, false, false, false);
        }
    }
    async hasNodeAccessRight(scope, node, ctx) {
        await this.loadRights(node, ctx);
        const rights = await this.getNodeAccessRights(node, ctx);
        return rights[scope];
    }
    async loadACLTrees(acl, ctx) {
        return ctx.accessRightsContexte.aclTrees.getOrSetValue(async () => {
            const aclType = await this.appCtx.accessRightsDefinitionType
                .waitForValue;
            const aclRoots = await this.objectNodeService.searchByParentIdsAndObjectTypeId(acl, aclType.id);
            const aclNodes = await this.objectNodeService.searchByTreeIds(aclRoots.map((aclRoot) => aclRoot.id));
            const aclTrees = {};
            for (const aclRoot of aclRoots) {
                aclTrees[aclRoot.parentNodeId] = await this.objectTreeService.buildTreeFromNodes(aclRoot, aclNodes, ctx);
            }
            return aclTrees;
        });
    }
    async loadRights(node, ctx) {
        return ctx.accessRightsContexte.rights.getOrSetValue(async () => {
            var _a, _b;
            const rights = new access_rights_const_1.AccessRightPermissions();
            try {
                const acl = node.aclList ? lodash_1.cloneDeep(node.aclList) : [];
                if (node.acl) {
                    acl.push(node.id);
                }
                if (0 === acl.length) {
                    ctx.accessRightsContexte.treeRootNodeId.value = null;
                    return rights;
                }
                ctx.accessRightsContexte.treeRootNodeId.value = acl[acl.length - 1];
                const aclTrees = await this.loadACLTrees(acl, ctx);
                const reset = new access_rights_const_1.AccessRightPermissions();
                if (node.id === acl[acl.length - 1] && node.id in aclTrees) {
                    const nodeRightSet = await this.computeACLTree(aclTrees[acl.pop()], ctx.accessRightsContexte.user.value);
                    reset.treeChildrenNodes = nodeRightSet.reset;
                    rights.treeChildrenNodes = nodeRightSet.rights;
                    (_a = rights.treeRootNode).read || (_a.read = rights.treeChildrenNodes.read);
                    (_b = rights.treeRootNode).create || (_b.create = rights.treeChildrenNodes.create);
                }
                let aclId = acl.pop();
                while (aclId &&
                    lodash_1.some(Object.keys(reset.treeChildrenNodes), (key) => false === reset.treeRootNode[key]) &&
                    lodash_1.some(Object.keys(rights.treeChildrenNodes), (key) => false === reset.treeRootNode[key])) {
                    if (aclId in aclTrees) {
                        const nodeRightSet = await this.computeACLTree(aclTrees[aclId], ctx.accessRightsContexte.user.value);
                        this.mergeRight(reset.treeChildrenNodes, rights.treeChildrenNodes, nodeRightSet.rights);
                        this.mergeReset(reset.treeChildrenNodes, nodeRightSet.reset);
                        this.mergeRight(reset.treeRootNode, rights.treeRootNode, nodeRightSet.rights);
                        this.mergeReset(reset.treeRootNode, nodeRightSet.reset);
                    }
                    aclId = acl.pop();
                }
                // eslint-disable-next-line no-empty
            }
            catch (error) { }
            return rights;
        });
    }
    mergeReset(actualResets, addedResets) {
        var _a;
        for (const permission in actualResets) {
            actualResets[_a = permission] || (actualResets[_a] = addedResets[permission]);
        }
    }
    mergeRight(resetedRights, actualRights, addedRights) {
        var _a;
        for (const right in resetedRights) {
            for (const permission in resetedRights) {
                if (!resetedRights[permission]) {
                    actualRights[_a = permission] || (actualRights[_a] = addedRights[permission]);
                }
            }
        }
    }
    async computeACLTree(aclTree, principal) {
        var _a, _b, _c, _d, _e;
        const computedTree = {
            reset: new access_rights_const_1.AccessRightSet(),
            rights: new access_rights_const_1.AccessRightSet(),
        };
        computedTree.reset.create = !!aclTree.treeNode.resetCreate;
        computedTree.reset.read = !!aclTree.treeNode.resetRead;
        computedTree.reset.update = !!aclTree.treeNode.resetUpdate;
        computedTree.reset.delete = !!aclTree.treeNode.resetDelete;
        const aclTreeSubTypes = aclTree.childrenByObjectTypeId;
        const ownersTypeId = this.appCtx.accessRightsOwnersType.value
            .id;
        if (ownersTypeId in aclTreeSubTypes) {
            const owners = aclTreeSubTypes[ownersTypeId][0];
            computedTree.reset.owner = !!owners.treeNode.resetOwner;
            if (this.findMatchingUser(owners.children, principal)) {
                computedTree.rights.owner = true;
                computedTree.rights.access = true;
                computedTree.rights.create = true;
                computedTree.rights.read = true;
                computedTree.rights.update = true;
                computedTree.rights.delete = true;
            }
        }
        const accessRightsAccessManagersId = this.appCtx
            .accessRightsAccessManagersType.value.id;
        if (accessRightsAccessManagersId in aclTreeSubTypes) {
            const accessRightsAccessManagers = aclTreeSubTypes[accessRightsAccessManagersId][0];
            computedTree.reset.access = !!accessRightsAccessManagers.treeNode
                .resetAccess;
            if (!computedTree.rights.access &&
                this.findMatchingUser(accessRightsAccessManagers.children, principal)) {
                computedTree.rights.access = true;
                computedTree.rights.create = true;
                computedTree.rights.read = true;
                computedTree.rights.update = true;
                computedTree.rights.delete = true;
            }
        }
        if (!(computedTree.rights.create &&
            computedTree.rights.read &&
            computedTree.rights.update &&
            computedTree.rights.delete)) {
            const groupTypeId = this.appCtx.accessRightsGroupType.value
                .id;
            if (groupTypeId in aclTreeSubTypes) {
                for (const group of aclTreeSubTypes[groupTypeId]) {
                    if (computedTree.rights.create &&
                        computedTree.rights.read &&
                        computedTree.rights.update &&
                        computedTree.rights.delete) {
                        break;
                    }
                    if ((!group.treeNode.grantCreate || computedTree.rights.create) &&
                        (!group.treeNode.grantRead || computedTree.rights.read) &&
                        (!group.treeNode.grantUpdate || computedTree.rights.update) &&
                        (!group.treeNode.grantDelete || computedTree.rights.delete)) {
                        continue;
                    }
                    if (this.findMatchingUser(group.children, principal)) {
                        (_a = computedTree.rights).create || (_a.create = group.treeNode.grantCreate);
                        (_b = computedTree.rights).read || (_b.read = group.treeNode.grantRead);
                        (_c = computedTree.rights).update || (_c.update = group.treeNode.grantUpdate);
                        (_d = computedTree.rights).delete || (_d.delete = group.treeNode.grantDelete);
                    }
                }
            }
            (_e = computedTree.rights).read || (_e.read = computedTree.rights.create ||
                computedTree.rights.update ||
                computedTree.rights.delete);
        }
        return computedTree;
    }
    findMatchingUser(users, principal) {
        return lodash_1.some(users, (user) => {
            switch (user.treeNode.objectTypeId) {
                case this.appCtx.userType.value.id:
                    return user.treeNode.contentUserId === principal.id;
                case this.appCtx.anonymousUserType.value.id:
                    return true;
            }
        });
    }
};
AccessRightsService = tslib_1.__decorate([
    core_1.injectable({ scope: core_1.BindingScope.SINGLETON }),
    tslib_1.__param(0, core_1.service(application_service_1.ApplicationService)),
    tslib_1.__param(1, core_1.service(object_type_service_1.ObjectTypeService)),
    tslib_1.__param(2, core_1.service(object_node_service_1.ObjectNodeService)),
    tslib_1.__param(3, core_1.service(object_tree_service_1.ObjectTreeService)),
    tslib_1.__metadata("design:paramtypes", [application_service_1.ApplicationService,
        object_type_service_1.ObjectTypeService,
        object_node_service_1.ObjectNodeService,
        object_tree_service_1.ObjectTreeService])
], AccessRightsService);
exports.AccessRightsService = AccessRightsService;
//# sourceMappingURL=access-rights.service.js.map